#!/usr/local/bin/python
import picos as pic
import networkx as nx
import cvxopt as cvx
import cvxopt.lapack
import numpy as np
import sys
import math
from pyomo.environ import *
from pyomo.opt import SolverStatus, TerminationCondition
import pyomo.opt
import time

GWitn = 10000
MyEpsilon = 1e-6
cplexTimeLimit = 100

## functions

def create_maxcut_miqp(G):
        # create max cut problem MIQP with PyOMO
        maxcut = ConcreteModel()
        # 0-1 vars: 0 if node in S, 1 if in \bar{S}
        maxcut.y = Var(G.nodes(), within = Binary)
        # objective: max sum_{ij in E} w_ij * (y_i + y_j - 2 y_i*y_j)
        def max_cut(model):
                return sum(G[i][j]['weight'] * (model.y[i] + model.y[j] - 2*model.y[i]*model.y[j])  for (i,j) in G.edges())
        maxcut.maxcutobj = Objective(rule = max_cut, sense = maximize)
        return maxcut

def create_maxcut_milp(G):
        # create max cut problem MILP with PyOMO
        maxcut = ConcreteModel()
        # 0-1 vars: 0 if node in S, 1 if in \bar{S}
        maxcut.y = Var(G.nodes(), within = Binary)
        # linearization variables z_ij = y_iy_j
        maxcut.z = Var(G.edges(), within = PercentFraction)
        # objective: max sum_{ij in E} w_ij * (y_i + y_j - 2 z_ij)
        def max_cut(model):
                return sum(G[i][j]['weight'] * (model.y[i] + model.y[j] - 2*model.z[i,j]) for (i,j) in G.edges())
        maxcut.maxcutobj = Objective(rule = max_cut, sense = maximize)
        # Fortet's linearization constraints for z = y*y
        def fortet1(model, i,j):
                return model.z[i,j] <= model.y[i]
        maxcut.fortet1 = Constraint(G.edges(), rule = fortet1)
        def fortet2(model, i,j):
                return model.z[i,j] <= model.y[j]
        maxcut.fortet2 = Constraint(G.edges(), rule = fortet2)
        def fortet3(model, i,j):
                return model.z[i,j] >= model.y[i] + model.y[j] - 1
        maxcut.fortet3 = Constraint(G.edges(), rule = fortet3)
        return maxcut
        
def create_maxcut_sdp(Laplacian):
        n = Laplacian.shape[0]
        # create max cut problem SDP with picos
        maxcut = pic.Problem()
        # SDP variable: n x n matrix
        X = maxcut.add_variable('X', (n,n), 'symmetric')
        # Laplacian of the graph (on the objective)
        #laplacian = 1/4.*nx.laplacian_matrix(G).todense()
        L = pic.new_param('L', Laplacian)
        # constraint: ones on the diagonal
        maxcut.add_constraint(pic.tools.diag_vect(X) == 1)
        # constraint: X positive semidefinite
        maxcut.add_constraint(X >> 0)
        # objective: A|B = tr(AB)
        maxcut.set_objective('max', L|X) 
        return maxcut

def rank_factor(X, rk):
        n = X.shape[0]
        evl, evc = np.linalg.eigh(X) # extract eigeninfo
        evl[evl < 0] = 0    # get closest nonneg eigenvls (some -1e-27 --> 0)
        if rk < n:
                x = np.transpose(evc[:,-rk:])
                for k in range(rk):
                        x[k] *= math.sqrt(evl[-rk-1])
        else:
                x = np.transpose(evc[::-1])
                for k in range(n):
                        x[k] *= math.sqrt(evl[::-1][k])
        return x

def closest_psd(X):
        evl, evc = np.linalg.eigh(X) # extract eigeninfo
        evl[evl < 0] = 0    # get closest nonneg eigenvls (some -1e-27 --> 0)
        return evc.T.dot(np.diag(evl).dot(evc))

def gw_randomized_rounding(V, objX, L, iterations):
        n = V.shape[0]
        # random projection algorithm
        # repeat max iterations times
        #   or, optionally, until within a factor .878 of SDP opt
        count = 0
        obj = 0
        print "maxcut(GW/rnd): improv_itn objfun"
        while (count < iterations):
                r = np.random.normal(0,1,n)  # rnd uniform vector on sphere:
                r /= np.linalg.norm(r)       #   optional (GW95)
                x = np.sign(np.dot(V,r))     # signs of V . r
                x[x >= 0] = 1                # in case there's some zeros
                o = np.dot(x.T,np.dot(L,x))  # objective function value
                if o > obj:
                        x_cut = x
                        obj = o
                        # if obj >= 0.878 * objX:
                        #         print count, "\t", obj, "\t [>= 0.878(OPT_sdp)]"
                        #         break
                        # else:
                        #         print count, "\t", obj
                        print count, "\t", obj
                count += 1
        return (x_cut,obj)

## main

# read edge list file as input
if len(sys.argv) < 2:
        print "error: need an edge list filename on cmd line"
        quit()        
G = nx.read_weighted_edgelist(sys.argv[1], nodetype = int) 

# make G undirected
G = nx.Graph(G)

## print edge weights
#for (i,j) in G.edges():
#        print G[i][j]['weight']

# solve MIQP using CPLEX
miqp = create_maxcut_miqp(G)
solver = pyomo.opt.SolverFactory('cplexamp', solver_io = 'nl')
solver.options['timelimit'] = cplexTimeLimit
solver.options['mipdisplay'] = 2
t0 = time.time()
results = solver.solve(miqp, keepfiles = False, tee = True)
t1 = time.time()
miqpcpu = t1 - t0
miqp.solutions.load_from(results)
ymiqp = {i:miqp.y[i].value for i in G.nodes()}
print 'maxcut(MIQP): solverstatus =', results.solver.status
miqpcut = [i for i,y in ymiqp.iteritems() if y == 1]
miqpobj = miqp.maxcutobj()
print "maxcut(MIQP): cut = ", miqpcut
print "maxcut(MIQP): value of cut = ", miqpobj

# solve MILP using CPLEX
milp = create_maxcut_milp(G)
#solver = pyomo.opt.SolverFactory('cplexamp', solver_io = 'nl')
#solver.options['timelimit'] = cplexTimeLimit
##solver.options['mipdisplay'] = 2
t0 = time.time()
results = solver.solve(milp, keepfiles = False, tee = True)
t1 = time.time()
milpcpu = t1 - t0
milp.solutions.load_from(results)
ymilp = {i:milp.y[i].value for i in G.nodes()}
print 'maxcut(MILP): solverstatus =', results.solver.status
milpcut = [i for i,y in ymilp.iteritems() if y == 1]
milpobj = milp.maxcutobj()
print "maxcut(MILP): cut = ", milpcut
print "maxcut(MILP): value of cut = ", milpobj

# solve SDP using picos
L = np.array(1/4.*nx.laplacian_matrix(G).todense())
sdp = create_maxcut_sdp(L)
t0 = time.time()
sdp.solve(verbose = 0)
t1 = time.time()
sdpcpu = t1 - t0
sdpobj = sdp.obj_value()
print 'maxcut(SDP): bound from SDP relaxation: {0}'.format(sdp.obj_value())
Xsdp = closest_psd(np.array(sdp.get_valued_variable('X')))

# GW rounding heuristic
t0 = time.time()
N = len(G.nodes())
V = rank_factor(Xsdp, N)
(gwcut,gwobj) = gw_randomized_rounding(V, sdp.obj_value(), L, GWitn)
t1 = time.time()
gwcpu = t1 - t0
S1 = [n+1 for n in range(N) if gwcut[n] >= 0]
S2 = [n+1 for n in range(N) if gwcut[n] < 0]
cutset = [(i,j) for (i,j) in G.edges() if (i in S1 and j in S2) or (i in S2 and j in S1)]
print 'maxcut(GW): cut = ', S1
print 'maxcut(GW): value of the cut = {0}'.format(gwobj)
if abs(gwobj - len(cutset)) > MyEpsilon:
       print 'maxcut(GW): WARNING: cut value = {0} != {1} = |cutset|'.format(gwobj, len(cutset))

# display statistics
print "maxcut(out): MIQPobj=", miqpobj, ", MIQPcpu=", miqpcpu
print "maxcut(out): MILPobj=", milpobj, ", MILPcpu=", milpcpu
print "maxcut(out):  SDPobj=",  sdpobj, ",  SDPcpu=",  sdpcpu
print "maxcut(out):   GWobj=",   gwobj, ",   GWcpu=",   gwcpu
